@@ -12,35 +12,23 @@
 #include <sys/types.h>
 #include <unistd.h>
 
-#include "libbcachefs/bcachefs.h"
-#include "libbcachefs/bcachefs_ioctl.h"
-#include "libbcachefs/errcode.h"
-#include "libbcachefs/journal.h"
-#include "libbcachefs/sb-members.h"
-#include "libbcachefs/super-io.h"
+#include <blkid.h>
+
 #include "cmds.h"
 #include "libbcachefs.h"
 #include "libbcachefs/opts.h"
 #include "tools-util.h"
 
-int device_usage(void)
-{
-       puts("bcachefs device - manage devices within a running filesystem\n"
-            "Usage: bcachefs device <CMD> [OPTION]\n"
-            "\n"
-            "Commands:\n"
-            "  add                     add a new device to an existing filesystem\n"
-            "  remove                  remove a device from an existing filesystem\n"
-            "  online                  re-add an existing member to a filesystem\n"
-            "  offline                 take a device offline, without removing it\n"
-            "  evacuate                migrate data off a specific device\n"
-            "  set-state               mark a device as failed\n"
-            "  resize                  resize filesystem on a device\n"
-            "  resize-journal          resize journal on a device\n"
-            "\n"
-            "Report bugs to <linux-bcachefs@vger.kernel.org>");
-       return 0;
-}
+#include "bcachefs.h"
+#include "bcachefs_ioctl.h"
+
+#include "init/dev.h"
+#include "init/fs.h"
+
+#include "journal/init.h"
+
+#include "sb/members.h"
+#include "sb/io.h"
 
 static void device_add_usage(void)
 {
@@ -51,14 +39,14 @@ static void device_add_usage(void)
 
 	bch2_opts_usage(OPT_FORMAT|OPT_DEVICE);
 
-	puts("  -l, --label=label           Disk label\n"
-	     "  -f, --force                 Use device even if it appears to already be formatted\n"
-	     "  -h, --help                  Display this help and exit\n"
+	puts("  -l, --label=label            Disk label\n"
+	     "  -f, --force                  Use device even if it appears to already be formatted\n"
+	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
-int cmd_device_add(int argc, char *argv[])
+static int cmd_device_add(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
 		{ "label",		required_argument,	NULL, 'l' },
@@ -66,7 +54,6 @@ int cmd_device_add(int argc, char *argv[])
 		{ "help",		no_argument,		NULL, 'h' },
 		{ NULL }
 	};
-	struct format_opts format_opts	= format_opts_default();
 	struct dev_opts dev_opts	= dev_opts_default();
 	bool force = false;
 
@@ -107,7 +94,7 @@ int cmd_device_add(int argc, char *argv[])
 			exit(EXIT_FAILURE);
 			break;
 		}
-}
+	}
 	args_shift(optind);
 
 	char *fs_path = arg_pop();
@@ -123,26 +110,26 @@ int cmd_device_add(int argc, char *argv[])
 
 	struct bchfs_handle fs = bcache_fs_open(fs_path);
 
-	int ret = open_for_format(&dev_opts, force);
+	int ret = open_for_format(&dev_opts, 0, force) ?:
+		bch2_format_for_device_add(&dev_opts,
+			read_file_u64(fs.sysfs_fd, "options/block_size"),
+			read_file_u64(fs.sysfs_fd, "options/btree_node_size"));
 	if (ret)
 		die("Error opening %s: %s", dev_opts.path, strerror(-ret));
 
-	struct bch_opt_strs fs_opt_strs;
-	memset(&fs_opt_strs, 0, sizeof(fs_opt_strs));
-
-	struct bch_opts fs_opts = bch2_parse_opts(fs_opt_strs);
-
-	opt_set(fs_opts, block_size,
-		read_file_u64(fs.sysfs_fd, "options/block_size"));
-	opt_set(fs_opts, btree_node_size,
-		read_file_u64(fs.sysfs_fd, "options/btree_node_size"));
-
-	struct bch_sb *sb = bch2_format(fs_opt_strs,
-					fs_opts,
-					format_opts,
-					&dev_opts, 1);
-	free(sb);
 	bchu_disk_add(fs, dev_opts.path);
+
+	/* A whole bunch of nonsense to get blkid to update its cache, so
+	 * mount can find the new device by UUID:
+	 */
+	blkid_cache cache = NULL;
+	if (!blkid_get_cache(&cache, NULL)) {
+		blkid_dev dev = blkid_get_dev(cache, dev_opts.path, BLKID_DEV_VERIFY);
+		if (dev)
+			blkid_verify(cache, dev);
+		blkid_put_cache(cache);
+	}
+
 	return 0;
 }
 
@@ -153,22 +140,21 @@ static void device_remove_usage(void)
 	     "  bcachefs device remove <device>|<devid> <path>\n"
 	     "\n"
 	     "Options:\n"
-	     "  -f, --force		    Force removal, even if some data\n"
-	     "                              couldn't be migrated\n"
-	     "  -F, --force-metadata	    Force removal, even if some metadata\n"
-	     "                              couldn't be migrated\n"
-	     "  -h, --help                  display this help and exit\n"
+	     "  -f, --force                  Force removal, even if some data couldn't be migrated\n"
+	     "  -F, --force-metadata         Force removal, even if some metadata couldn't be migrated\n"
+	     "  -h, --help                   Display this help and exit\n"
+	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 	exit(EXIT_SUCCESS);
 }
 
-int cmd_device_remove(int argc, char *argv[])
+static int cmd_device_remove(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "by-id",              0, NULL, 'i' },
-		{ "force",		0, NULL, 'f' },
-		{ "force-metadata",	0, NULL, 'F' },
-		{ "help",		0, NULL, 'h' },
+		{ "by-id",              no_argument, NULL, 'i' },
+		{ "force",		no_argument, NULL, 'f' },
+		{ "force-metadata",	no_argument, NULL, 'F' },
+		{ "help",		no_argument, NULL, 'h' },
 		{ NULL }
 	};
 	struct bchfs_handle fs;
@@ -223,16 +209,20 @@ static void device_online_usage(void)
 	     "Usage: bcachefs device online [OPTION]... device\n"
 	     "\n"
 	     "Options:\n"
-	     "  -h, --help                  Display this help and exit\n"
+	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
-int cmd_device_online(int argc, char *argv[])
+static int cmd_device_online(int argc, char *argv[])
 {
+	static const struct option longopts[] = {
+		{ "help",		no_argument, NULL, 'h' },
+		{ NULL }
+	};
 	int opt;
 
-	while ((opt = getopt(argc, argv, "h")) != -1)
+	while ((opt = getopt_long(argc, argv, "h", longopts, NULL)) != -1)
 		switch (opt) {
 		case 'h':
 			device_online_usage();
@@ -259,16 +249,17 @@ static void device_offline_usage(void)
 	     "Usage: bcachefs device offline [OPTION]... device\n"
 	     "\n"
 	     "Options:\n"
-	     "  -f, --force		    Force, if data redundancy will be degraded\n"
-	     "  -h, --help                  Display this help and exit\n"
+	     "  -f, --force                  Force, if data redundancy will be degraded\n"
+	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
-int cmd_device_offline(int argc, char *argv[])
+static int cmd_device_offline(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "force",		0, NULL, 'f' },
+		{ "force",		no_argument, NULL, 'f' },
+		{ "help",		no_argument, NULL, 'h' },
 		{ NULL }
 	};
 	int opt, flags = 0;
@@ -304,17 +295,26 @@ static void device_evacuate_usage(void)
 	     "Usage: bcachefs device evacuate [OPTION]... device\n"
 	     "\n"
 	     "Options:\n"
-	     "  -h, --help                  Display this help and exit\n"
+	     "  -f, --force                  Force if data redundancy will be degraded\n"
+	     "  -h, --help                   Display this help and exit\n"
 	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 }
 
-int cmd_device_evacuate(int argc, char *argv[])
+static int cmd_device_evacuate(int argc, char *argv[])
 {
-	int opt;
+	static const struct option longopts[] = {
+		{ "force",			no_argument,	NULL, 'f' },
+		{ "help",			no_argument,	NULL, 'h' },
+		{ NULL }
+	};
+	int opt, flags = 0;
 
-	while ((opt = getopt(argc, argv, "h")) != -1)
+	while ((opt = getopt_long(argc, argv, "fh", longopts, NULL)) != -1)
 		switch (opt) {
+		case 'f':
+			flags |= BCH_FORCE_IF_DEGRADED;
+			break;
 		case 'h':
 			device_evacuate_usage();
 			exit(EXIT_SUCCESS);
@@ -335,7 +335,7 @@ int cmd_device_evacuate(int argc, char *argv[])
 
 	if (u->state == BCH_MEMBER_STATE_rw) {
 		printf("Setting %s readonly\n", dev_path);
-		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, 0);
+		bchu_disk_set_state(fs, dev_idx, BCH_MEMBER_STATE_ro, flags);
 	}
 
 	free(u);
@@ -359,21 +359,22 @@ static void device_set_state_usage(void)
 	     "<path>: path to mounted filesystem, optional unless specifying device by id\n"
 	     "\n"
 	     "Options:\n"
-	     "  -f, --force		    Force, if data redundancy will be degraded\n"
-	     "      --force-if-data-lost    Force, if data will be lost\n"
-	     "  -o, --offline               Set state of an offline device\n"
-	     "  -h, --help                  display this help and exit\n"
+	     "  -f, --force                  Force if data redundancy will be degraded\n"
+	     "      --force-if-data-lost     Force if data will be lost\n"
+	     "  -o, --offline                Set state of an offline device\n"
+	     "  -h, --help                   Display this help and exit\n"
+	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 	exit(EXIT_SUCCESS);
 }
 
-int cmd_device_set_state(int argc, char *argv[])
+static int cmd_device_set_state(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "force",			0, NULL, 'f' },
-		{ "force-if-data-lost",		0, NULL, 'F' },
-		{ "offline",			0, NULL, 'o' },
-		{ "help",			0, NULL, 'h' },
+		{ "force",			no_argument,	NULL, 'f' },
+		{ "force-if-data-lost",		no_argument,	NULL, 'F' },
+		{ "offline",			no_argument,	NULL, 'o' },
+		{ "help",			no_argument,	NULL, 'h' },
 		{ NULL }
 	};
 	struct bchfs_handle fs;
@@ -415,27 +416,37 @@ int cmd_device_set_state(int argc, char *argv[])
 		by_id = true;
 
 	if (offline) {
-		struct bch_opts opts = bch2_opts_empty();
-		struct bch_sb_handle sb = { NULL };
-
 		if (by_id)
 			die("Cannot specify offline device by id");
 
+		struct bch_opts opts = bch2_opts_empty();
+		opt_set(opts, nostart,	true);
+		opt_set(opts, degraded, BCH_DEGRADED_very);
+
+		struct bch_sb_handle sb = { NULL };
 		int ret = bch2_read_super(dev_str, &opts, &sb);
 		if (ret)
 			die("error opening %s: %s", dev_str, bch2_err_str(ret));
 
-		struct bch_member *m = bch2_members_v2_get_mut(sb.sb, sb.sb->dev_idx);
-
-		SET_BCH_MEMBER_STATE(m, new_state);
-
-		le64_add_cpu(&sb.sb->seq, 1);
-
-		bch2_super_write(sb.bdev->bd_fd, sb.sb);
-		ret = fsync(sb.bdev->bd_fd);
-		if (ret)
-			fprintf(stderr, "error writing superblock: fsync error (%m)");
+		unsigned dev_idx = sb.sb->dev_idx;
 		bch2_free_super(&sb);
+
+		/* scan for all devices in fs */
+		darray_const_str devs = get_or_split_cmdline_devs(1, &dev_str);
+
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
+		ret = PTR_ERR_OR_ZERO(c);
+		if (ret)
+			die("Error opening filesystem: %s", bch2_err_str(ret));
+
+		scoped_guard(mutex, &c->sb_lock) {
+			struct bch_member *m = bch2_members_v2_get_mut(c->disk_sb.sb, dev_idx);
+
+			SET_BCH_MEMBER_STATE(m, new_state);
+
+			bch2_write_super(c);
+		}
+		bch2_fs_stop(c);
 		return ret;
 	}
 
@@ -466,15 +477,16 @@ static void device_resize_usage(void)
 	     "Usage: bcachefs device resize device [ size ]\n"
 	     "\n"
 	     "Options:\n"
-	     "  -h, --help                  display this help and exit\n"
+	     "  -h, --help                   Display this help and exit\n"
+	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 	exit(EXIT_SUCCESS);
 }
 
-int cmd_device_resize(int argc, char *argv[])
+static int cmd_device_resize(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "help",			0, NULL, 'h' },
+		{ "help",			no_argument, NULL, 'h' },
 		{ NULL }
 	};
 	u64 size;
@@ -523,7 +535,7 @@ int cmd_device_resize(int argc, char *argv[])
 
 		struct bch_member m = bch2_sb_member_get(sb, idx);
 
-		u64 nbuckets = size / le16_to_cpu(m.bucket_size);
+		u64 nbuckets = size / BCH_MEMBER_BUCKET_SIZE(&m);
 
 		if (nbuckets < le64_to_cpu(m.nbuckets))
 			die("Shrinking not supported yet");
@@ -533,30 +545,35 @@ int cmd_device_resize(int argc, char *argv[])
 	} else {
 		printf("Doing offline resize of %s\n", dev);
 
-		struct bch_fs *c = bch2_fs_open(&dev, 1, bch2_opts_empty());
+		darray_const_str devs = {};
+		darray_push(&devs, dev);
+
+		struct bch_opts opts = bch2_opts_empty();
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
 		if (IS_ERR(c))
 			die("error opening %s: %s", dev, bch2_err_str(PTR_ERR(c)));
 
 		struct bch_dev *resize = NULL;
 
-		for_each_online_member(c, ca) {
+		for_each_online_member(c, ca, 0) {
 			if (resize)
 				die("confused: more than one online device?");
 			resize = ca;
-			percpu_ref_get(&resize->io_ref);
+			enumerated_ref_get(&resize->io_ref[READ], 0);
 		}
 
-		u64 nbuckets = size / le16_to_cpu(resize->mi.bucket_size);
+		u64 nbuckets = size / resize->mi.bucket_size;
 
 		if (nbuckets < le64_to_cpu(resize->mi.nbuckets))
 			die("Shrinking not supported yet");
 
 		printf("resizing %s to %llu buckets\n", dev, nbuckets);
-		int ret = bch2_dev_resize(c, resize, nbuckets);
+		CLASS(printbuf, err)();
+		int ret = bch2_dev_resize(c, resize, nbuckets, &err);
 		if (ret)
-			fprintf(stderr, "resize error: %s\n", bch2_err_str(ret));
+			fprintf(stderr, "resize error: %s\n%s", bch2_err_str(ret), err.buf);
 
-		percpu_ref_put(&resize->io_ref);
+		enumerated_ref_put(&resize->io_ref[READ], 0);
 		bch2_fs_stop(c);
 	}
 	return 0;
@@ -568,15 +585,16 @@ static void device_resize_journal_usage(void)
 	     "Usage: bcachefs device resize-journal device size\n"
 	     "\n"
 	     "Options:\n"
-	     "  -h, --help                  display this help and exit\n"
+	     "  -h, --help                   Display this help and exit\n"
+	     "\n"
 	     "Report bugs to <linux-bcachefs@vger.kernel.org>");
 	exit(EXIT_SUCCESS);
 }
 
-int cmd_device_resize_journal(int argc, char *argv[])
+static int cmd_device_resize_journal(int argc, char *argv[])
 {
 	static const struct option longopts[] = {
-		{ "help",			0, NULL, 'h' },
+		{ "help",			no_argument, NULL, 'h' },
 		{ NULL }
 	};
 	u64 size;
@@ -623,24 +641,28 @@ int cmd_device_resize_journal(int argc, char *argv[])
 
 		struct bch_member m = bch2_sb_member_get(sb, idx);
 
-		u64 nbuckets = size / le16_to_cpu(m.bucket_size);
+		u64 nbuckets = size / BCH_MEMBER_BUCKET_SIZE(&m);
 
 		printf("resizing journal on %s to %llu buckets\n", dev, nbuckets);
 		bchu_disk_resize_journal(fs, idx, nbuckets);
 	} else {
 		printf("%s is offline - starting:\n", dev);
 
-		struct bch_fs *c = bch2_fs_open(&dev, 1, bch2_opts_empty());
+		darray_const_str devs = {};
+		darray_push(&devs, dev);
+
+		struct bch_opts opts = bch2_opts_empty();
+		struct bch_fs *c = bch2_fs_open(&devs, &opts);
 		if (IS_ERR(c))
 			die("error opening %s: %s", dev, bch2_err_str(PTR_ERR(c)));
 
 		struct bch_dev *resize = NULL;
 
-		for_each_online_member(c, ca) {
+		for_each_online_member(c, ca, 0) {
 			if (resize)
 				die("confused: more than one online device?");
 			resize = ca;
-			percpu_ref_get(&resize->io_ref);
+			enumerated_ref_get(&resize->io_ref[READ], 0);
 		}
 
 		u64 nbuckets = size / le16_to_cpu(resize->mi.bucket_size);
@@ -650,8 +672,54 @@ int cmd_device_resize_journal(int argc, char *argv[])
 		if (ret)
 			fprintf(stderr, "resize error: %s\n", bch2_err_str(ret));
 
-		percpu_ref_put(&resize->io_ref);
+		enumerated_ref_put(&resize->io_ref[READ], 0);
 		bch2_fs_stop(c);
 	}
 	return 0;
 }
+
+static int device_usage(void)
+{
+       puts("bcachefs device - manage devices within a running filesystem\n"
+            "Usage: bcachefs device <CMD> [OPTION]\n"
+            "\n"
+            "Commands:\n"
+            "  add                          Add a new device to an existing filesystem\n"
+            "  remove                       Remove a device from an existing filesystem\n"
+            "  online                       Re-add an existing member to a filesystem\n"
+            "  offline                      Take a device offline, without removing it\n"
+            "  evacuate                     Migrate data off a specific device\n"
+            "  set-state                    Mark a device as failed\n"
+            "  resize                       Resize filesystem on a device\n"
+            "  resize-journal               Resize journal on a device\n"
+            "\n"
+            "Report bugs to <linux-bcachefs@vger.kernel.org>");
+       return 0;
+}
+
+int device_cmds(int argc, char *argv[])
+{
+	char *cmd = pop_cmd(&argc, argv);
+
+	if (argc < 1)
+		return device_usage();
+	if (!strcmp(cmd, "add"))
+		return cmd_device_add(argc, argv);
+	if (!strcmp(cmd, "remove"))
+		return cmd_device_remove(argc, argv);
+	if (!strcmp(cmd, "online"))
+		return cmd_device_online(argc, argv);
+	if (!strcmp(cmd, "offline"))
+		return cmd_device_offline(argc, argv);
+	if (!strcmp(cmd, "evacuate"))
+		return cmd_device_evacuate(argc, argv);
+	if (!strcmp(cmd, "set-state"))
+		return cmd_device_set_state(argc, argv);
+	if (!strcmp(cmd, "resize"))
+		return cmd_device_resize(argc, argv);
+	if (!strcmp(cmd, "resize-journal"))
+		return cmd_device_resize_journal(argc, argv);
+
+	device_usage();
+	return -EINVAL;
+}
