@@ -555,31 +555,23 @@ void *__bch2_trans_subbuf_alloc(struct btree_trans *trans,
 }
 
 int bch2_bkey_get_empty_slot(struct btree_trans *trans, struct btree_iter *iter,
-			     enum btree_id btree, struct bpos end)
+			     enum btree_id btree, struct bpos start, struct bpos end)
 {
 	bch2_trans_iter_init(trans, iter, btree, end, BTREE_ITER_intent);
-	struct bkey_s_c k = bch2_btree_iter_peek_prev(iter);
-	int ret = bkey_err(k);
-	if (ret)
-		goto err;
+	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_prev(iter));
 
-	bch2_btree_iter_advance(iter);
-	k = bch2_btree_iter_peek_slot(iter);
-	ret = bkey_err(k);
-	if (ret)
-		goto err;
+	if (bpos_lt(iter->pos, start))
+		bch2_btree_iter_set_pos(iter, start);
+	else
+		bch2_btree_iter_advance(iter);
 
+	k = bkey_try(bch2_btree_iter_peek_slot(iter));
 	BUG_ON(k.k->type != KEY_TYPE_deleted);
 
-	if (bkey_gt(k.k->p, end)) {
-		ret = bch_err_throw(trans->c, ENOSPC_btree_slot);
-		goto err;
-	}
+	if (bkey_gt(k.k->p, end))
+		return bch_err_throw(trans->c, ENOSPC_btree_slot);
 
 	return 0;
-err:
-	bch2_trans_iter_exit(iter);
-	return ret;
 }
 
 void bch2_trans_commit_hook(struct btree_trans *trans,
@@ -653,54 +645,53 @@ int bch2_btree_delete(struct btree_trans *trans,
 		bch2_btree_delete_at(trans, &iter, flags);
 }
 
+static int delete_range_one(struct btree_trans *trans, struct btree_iter *iter,
+			    struct bpos end, enum btree_iter_update_trigger_flags flags)
+{
+	struct bkey_s_c k = bkey_try(bch2_btree_iter_peek_max(iter, end));
+
+	if (!k.k)
+		return 1;
+
+	CLASS(disk_reservation, res)(trans->c);
+
+	/*
+	 * This could probably be more efficient for extents:
+	 *
+	 * For extents, iter.pos won't necessarily be the same as
+	 * bkey_start_pos(k.k) (for non extents they always will be the
+	 * same). It's important that we delete starting from iter.pos
+	 * because the range we want to delete could start in the middle
+	 * of k.
+	 *
+	 * (bch2_btree_iter_peek() does guarantee that iter.pos >=
+	 * bkey_start_pos(k.k)).
+	 */
+	struct bkey_i delete;
+	bkey_init(&delete.k);
+	delete.k.p = iter->pos;
+
+	if (iter->flags & BTREE_ITER_is_extents)
+		bch2_key_resize(&delete.k,
+				bpos_min(end, k.k->p).offset -
+				iter->pos.offset);
+
+	try(bch2_trans_update(trans, iter, &delete, flags));
+	try(bch2_trans_commit(trans, &res.r, NULL, BCH_TRANS_COMMIT_no_enospc));
+	return 0;
+}
+
 int bch2_btree_delete_range_trans(struct btree_trans *trans, enum btree_id btree,
 				  struct bpos start, struct bpos end,
-				  enum btree_iter_update_trigger_flags flags,
-				  u64 *journal_seq)
+				  enum btree_iter_update_trigger_flags flags)
 {
 	u32 restart_count = trans->restart_count;
-	struct bkey_s_c k;
 	int ret = 0;
 
 	CLASS(btree_iter, iter)(trans, btree, start, BTREE_ITER_intent|flags);
 
-	while ((k = bch2_btree_iter_peek_max(&iter, end)).k) {
-		struct disk_reservation disk_res =
-			bch2_disk_reservation_init(trans->c, 0);
-		struct bkey_i delete;
-
-		ret = bkey_err(k);
-		if (ret)
-			goto err;
-
-		bkey_init(&delete.k);
-
-		/*
-		 * This could probably be more efficient for extents:
-		 */
-
-		/*
-		 * For extents, iter.pos won't necessarily be the same as
-		 * bkey_start_pos(k.k) (for non extents they always will be the
-		 * same). It's important that we delete starting from iter.pos
-		 * because the range we want to delete could start in the middle
-		 * of k.
-		 *
-		 * (bch2_btree_iter_peek() does guarantee that iter.pos >=
-		 * bkey_start_pos(k.k)).
-		 */
-		delete.k.p = iter.pos;
-
-		if (iter.flags & BTREE_ITER_is_extents)
-			bch2_key_resize(&delete.k,
-					bpos_min(end, k.k->p).offset -
-					iter.pos.offset);
-
-		ret   = bch2_trans_update(trans, &iter, &delete, flags) ?:
-			bch2_trans_commit(trans, &disk_res, journal_seq,
-					  BCH_TRANS_COMMIT_no_enospc);
-		bch2_disk_reservation_put(trans->c, &disk_res);
-err:
+	while (true) {
+		ret = delete_range_one(trans, &iter, end, flags);
 		/*
 		 * the bch2_trans_begin() call is in a weird place because we
 		 * need to call it after every transaction commit, to avoid path
@@ -715,7 +706,7 @@ err:
 			break;
 	}
 
-	return ret ?: trans_was_restarted(trans, restart_count);
+	return ret < 0 ? ret : trans_was_restarted(trans, restart_count);
 }
 
 /*
@@ -725,11 +716,10 @@ err:
  */
 int bch2_btree_delete_range(struct bch_fs *c, enum btree_id id,
 			    struct bpos start, struct bpos end,
-			    enum btree_iter_update_trigger_flags flags,
-			    u64 *journal_seq)
+			    enum btree_iter_update_trigger_flags flags)
 {
 	CLASS(btree_trans, trans)(c);
-	int ret = bch2_btree_delete_range_trans(trans, id, start, end, flags, journal_seq);
+	int ret = bch2_btree_delete_range_trans(trans, id, start, end, flags);
 	if (ret == -BCH_ERR_transaction_restart_nested)
 		ret = 0;
 	return ret;
