@@ -6,6 +6,7 @@
 #include "bcachefs.h"
 #include "opts.h"
 
+#include "alloc/background.h"
 #include "alloc/disk_groups.h"
 
 #include "data/compress.h"
@@ -546,6 +547,7 @@ int bch2_opt_hook_pre_set(struct bch_fs *c, struct bch_dev *ca, u64 inum, enum b
 	}
 
 	if (change &&
+	    test_bit(BCH_FS_started, &c->flags) &&
 	    (id == Opt_foreground_target ||
 	     id == Opt_background_target ||
 	     id == Opt_promote_target ||
@@ -569,13 +571,14 @@ int bch2_opts_hooks_pre_set(struct bch_fs *c)
 void bch2_opt_hook_post_set(struct bch_fs *c, struct bch_dev *ca, u64 inum,
 			    enum bch_opt_id id, u64 v)
 {
-	if (id == Opt_foreground_target ||
-	    id == Opt_background_target ||
-	    id == Opt_promote_target ||
-	    id == Opt_compression ||
-	    id == Opt_background_compression ||
-	    id == Opt_data_checksum ||
-	    id == Opt_data_replicas) {
+	if (test_bit(BCH_FS_started, &c->flags) &&
+	    (id == Opt_foreground_target ||
+	     id == Opt_background_target ||
+	     id == Opt_promote_target ||
+	     id == Opt_compression ||
+	     id == Opt_background_compression ||
+	     id == Opt_data_checksum ||
+	     id == Opt_data_replicas)) {
 		bch2_set_rebalance_needs_scan(c, inum);
 		bch2_rebalance_wakeup(c);
 	}
@@ -599,6 +602,15 @@ void bch2_opt_hook_post_set(struct bch_fs *c, struct bch_dev *ca, u64 inum,
 			bch2_write_super(c);
 		}
 		break;
+	case Opt_durability:
+		if (test_bit(BCH_FS_rw, &c->flags) &&
+		    ca &&
+		    bch2_dev_is_online(ca) &&
+		    ca->mi.state == BCH_MEMBER_STATE_rw) {
+			guard(rcu)();
+			bch2_dev_allocator_set_rw(c, ca, true);
+		}
+		break;
 	case Opt_version_upgrade:
 		/*
 		 * XXX: in the future we'll likely want to do compatible
